# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
import os
import shutil
import datetime

from vcstools import VcsClient
from vcstools.vcs_base import VcsError

from common import MultiProjectException
from config_yaml import PathSpec
import ui

# helper class

class PreparationReport:
  """Specifies after user interaction of how to perform install / update"""
  def __init__(self, element):
    self.config_element = element
    self.abort = False # abort ALL operations
    self.skip = False # skip this tree
    self.error = None # message
    self.verbose = False # verbosity
    self.checkout = True # checkout vs update
    self.backup = False # backup vs delete
    self.backup_path = None # where to move tree to

## Each Config element provides actions on a local folder
    
class ConfigElement:
  """ Base class for Config provides methods with not implemented
  exceptions.  Also a few shared methods."""
  def __init__(self, path, local_name, properties = None):
    self.path = path
    if path is None:
      raise MultiProjectException("Invalid empty path")
    self.local_name = local_name
    self.properties = properties
  def get_path(self):
    """A normalized absolute path"""
    return self.path
  def get_local_name(self):
    """What the user specified in his config"""
    return self.local_name
  def prepare_install(self, backup_path = None, arg_mode = 'abort', robust = False):
    """
    Check whether install can be performed, asking user for decision if necessary.
    
    :param arg_mode: one of prompt, backup, delete, skip. Determines how to handle error cases
    :param backup_path: if arg_mode==backup, determines where to backup to
    :param robust: if true, operation will be aborted without changes to the filesystem and without user interaction
    :returns: A preparation_report instance, telling whether to checkout or to update, how to deal with existing tree, and where to backup to.
    """
    preparation_report = PreparationReport(self)
    preparation_report.skip = True
    return preparation_report
  
  def install(self, checkout = True, backup = False, backup_path = None, verbose = False):
    """
    Attempt to make it so that self.path is the result of checking out / updating from remote repo.
    No user Interaction allowed here (for concurrent mode).
    
    :param checkout: whether to checkout or update
    :param backup: if checking out, what to do if path exists. If true, backup_path must be set.
    """
    raise NotImplementedError, "ConfigElement install unimplemented"
  
  def get_path_spec(self):
    """PathSpec object with values as specified in file"""
    raise NotImplementedError, "ConfigElement get_path_spec unimplemented"
  def get_properties(self):
    """Any meta information attached"""
    return self.properties
  def get_versioned_path_spec(self):
    """PathSpec where VCS elements have the version looked up"""
    raise NotImplementedError, "ConfigElement get_versioned_path_spec unimplemented"
  def is_vcs_element(self):
    # subclasses to override when appropriate
    return False
  def get_diff(self, basepath = None):
    raise NotImplementedError, "ConfigElement get_diff unimplemented"
  def get_status(self, basepath = None, untracked = False):
    raise NotImplementedError, "ConfigElement get_status unimplemented"
  def backup(self, backup_path):
    if not backup_path:
      raise MultiProjectException("[%s] Cannot install %s.  backup disabled."%(self.get_local_name(), self.get_path()))
    backup_path = os.path.join(backup_path, os.path.basename(self.path)+"_%s"%datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S"))
    print("[%s] Backing up %s to %s"%(self.get_local_name(), self.get_path(), backup_path))
    shutil.move(self.path, backup_path)
  def __str__(self):
    return str(self.get_path_spec().get_legacy_yaml());
  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.get_path_spec() == other.get_path_spec()
    else:
      return False


class OtherConfigElement(ConfigElement):

  def __init__(self, path, local_name, uri=None, version='', properties = None):
    ConfigElement.__init__(self, path, local_name, properties)
    self.uri = uri
    self.version = version

  
  def install(self, checkout = True, backup = False, backup_path = None, verbose = False):
    return True

  def get_versioned_path_spec(self):
    raise MultiProjectException("Cannot generate versioned outputs with non source types")

  def get_path_spec(self):
    "yaml as from source"
    version = self.version
    if version == '': version = None
    return PathSpec(local_name = self.get_local_name(),
                    path = self.get_path(),
                    scmtype = None,
                    uri = self.uri,
                    version = version,
                    tags = self.get_properties())


class SetupConfigElement(ConfigElement):
  """A setup config element specifies a single file containing configuration data for a config."""

  def install(self, checkout = True, backup = False, backup_path = None, verbose = False):
    return True

  def get_versioned_path_spec(self):
    raise MultiProjectException("Cannot generate versioned outputs with non source types")
  
  def get_path_spec(self):
    return PathSpec(local_name = self.get_local_name(), path = self.get_path(), tags=['setup-file'] + (self.get_properties() or []))


class VCSConfigElement(ConfigElement):
  
  def __init__(self, path, local_name, uri, version='', properties = None):
    """
    Creates a config element for a VCS repository.
    
    :param path: absolute or relative path, str
    :param vcs_client: Object compatible with vcstools.VcsClientBase
    :param local_name: display name for the element, str
    :param uri: VCS uri to checkout/pull from, str
    :param version: optional revision spec (tagname, SHAID, ..., str)
    """
    ConfigElement.__init__(self, path, local_name, properties)
    if uri is None:
      raise MultiProjectException("Invalid scm entry having no uri attribute for path %s"%path)
    self.uri = uri.rstrip('/') # strip trailing slashes if defined to not be too strict #3061
    self.version = version

  def _get_vcsc(self):
    raise NotImplementedError, "VCSConfigElement _get_vcsc() unimplemented"
  
  def is_vcs_element(self):
    return True

  def get_vcs_type_name(self):
    # also see override in AVCSConfigElement
    return self._get_vcsc().get_vcs_type_name()

  def detect_presence(self):
    # also see override in AVCSConfigElement
    return self._get_vcsc().detect_presence()

  def path_exists(self):
    # we could use _get_vcsc().path_exit(), but this causes a time penalty for initializing
    # this is crucial for bash tab completion
    return os.path.isdir(self.path)
  
  def prepare_install(self, backup_path = None, arg_mode = 'abort', robust = False):
    preparation_report = PreparationReport(self)
    present = self.detect_presence()
    if present or self.path_exists():
      # Directory exists see what we need to do
      error_message = None
      
      if not present:
        error_message = "Failed to detect %s presence at %s."%(self.get_vcs_type_name(), self.path)
      else:
        cur_url = self._get_vcsc().get_url()
        if cur_url is not None:
          cur_url = cur_url.rstrip('/')  #strip trailing slashes for #3269
        if not cur_url or cur_url != self.uri.rstrip('/'):
          # local repositories get absolute pathnames
          if not (os.path.isdir(self.uri) and os.path.isdir(cur_url) and os.path.samefile(cur_url, self.uri)):
            error_message = "Url %s does not match %s requested."%(cur_url, self.uri)
      if error_message is None:
        # update should be possible
        preparation_report.checkout = False
      else:
        # If robust ala continue-on-error, just error now and it will be continued at a higher level
        if robust:
          raise MultiProjectException("Update Failed of %s: %s"%(self.path, error_message))
        # prompt the user based on the error code
        if arg_mode == 'prompt':
          print("Prepare updating %s (%s) to %s"%(self.uri, self.version, self.path))
          mode = ui.Ui.get_ui().prompt_del_abort_retry(error_message, allow_skip = True)
        else:
          mode = arg_mode
        if mode == 'backup':
            preparation_report.backup = True
            if backup_path == None:
              print("Prepare updating %s (%s) to %s"%(self.uri, self.version, self.path))
              preparation_report.backup_path = ui.Ui.get_ui().get_backup_path()
            else:
              preparation_report.backup_path = backup_path
        if mode == 'abort':
          preparation_report.abort = True
          preparation_report.error = error_message
        if mode == 'skip':
          preparation_report.skip = True
          preparation_report.error = error_message
        if mode == 'delete':
          preparation_report.backup = False
    return preparation_report

  def install(self, checkout = True, backup = True, backup_path = None, verbose = False):
    """
    Runs the equivalent of SCM checkout for new local repos or update for existing.
    
    :param checkout: whether to use an update command or a checkout/clone command
    :param backup: if checkout is True and folder exists, if backup is false folder will be DELETED.
    :param backup_path: if checkout is true and backup is true, move folder to this location
    """
    if checkout is True:
      print("[%s] Installing %s (%s) to %s"%(self.get_local_name(), self.uri, self.version, self.get_path()))
      if self.path_exists():
        if (backup is False):
          shutil.rmtree(self.path)
        else:
          self.backup(backup_path)
      if not self._get_vcsc().checkout(self.uri, self.version, verbose = verbose):
        raise MultiProjectException("[%s] Checkout of %s version %s into %s failed."%(self.get_local_name(), self.uri, self.version, self.get_path()))
    else:
      print("[%s] Updating %s"%(self.get_local_name(), self.get_path()))
      if not self._get_vcsc().update(self.version, verbose = verbose):
        raise MultiProjectException("[%s] Update Failed of %s"%(self.get_local_name(), self.get_path()))
    print("[%s] Done."%self.get_local_name())
          
  def get_path_spec(self):
    "yaml as from source"
    version = self.version
    if version == '': version = None
    return PathSpec(local_name = self.get_local_name(),
                    path = self.get_path(),
                    scmtype = self.get_vcs_type_name(),
                    uri = self.uri,
                    version = version,
                    tags = self.get_properties())

  def get_versioned_path_spec(self):
    "yaml looking up current version"
    version = self.version
    if version == '': version = None
    revision = None
    if version is not None:
      # revision is the UID of the version spec, can be them same
      revision = self._get_vcsc().get_version(self.version)
    currevision = self._get_vcsc().get_version()
    return PathSpec(local_name = self.get_local_name(),
                    path = self.get_path(),
                    scmtype = self.get_vcs_type_name(),
                    uri = self.uri,
                    version = version,
                    revision = revision,
                    currevision = currevision,
                    curr_uri = self._get_vcsc().get_url(),
                    tags = self.get_properties())
     

  def get_diff(self, basepath=None):
    return self._get_vcsc().get_diff(basepath)
  
  def get_status(self, basepath=None, untracked=False):
    return self._get_vcsc().get_status(basepath, untracked)
  

  
class AVCSConfigElement(VCSConfigElement):
  """
  Implementation using vcstools vcsclient, works for types svn, git, hg, bzr, tar
  
  :raises: Lookup Exception for unknown types
  """
  def __init__(self, scmtype, path, local_name, uri, version = '', vcsc = None, properties = None):
    VCSConfigElement.__init__(self, path, local_name = local_name, uri = uri, version = version, properties = properties)
    self.vcsc = vcsc
    self._scmtype = scmtype

  def get_vcs_type_name(self):
    return self._scmtype
    
  def _get_vcsc(self):
    # lazy initializer
    if self.vcsc == None:
      try:
        self.vcsc = VcsClient(self._scmtype, self.get_path())
      except VcsError as e:
        raise MultiProjectException("Unable to create vcs client of type %s for %s: %s"%(self._scmtype, self.get_path(), e))
    return self.vcsc

  def detect_presence(self):
    # to make more use of lazy initializer, do not instantiate client for just this
    # this is crucial for bash tab completion
    if self.get_vcs_type_name() == 'git':
      return os.path.isdir(os.path.join(self.path, '.git'))
    elif self.get_vcs_type_name() == 'svn':
      return os.path.isdir(os.path.join(self.path, '.svn'))
    elif self.get_vcs_type_name() == 'hg':
      return os.path.isdir(os.path.join(self.path, '.hg'))
    elif self.get_vcs_type_name() == 'bzr':
      return os.path.isdir(os.path.join(self.path, '.bzr'))
    else:
      return self._get_vcsc().detect_presence()



from __future__ import print_function
import os
import sys
import distutils
from subprocess import Popen, PIPE
from optparse import OptionParser

import yaml

from rosinstall.helpers import ROSInstallException, ROSINSTALL_FILENAME
import rosinstall.helpers
import rosinstall.cli_common
import rosinstall.rosws_cli
import rosinstall.rosinstall_cmd
import rosinstall.multiproject_cmd
import rosinstall.config_yaml



def get_stack_element_in_config(config, stack):
  """
  The path_spec of a config element if it is named like stack and has a
  root level file named stack.xml
  """
  for entry in config.get_config_elements():
    if entry.get_local_name() == stack:
      if os.path.isfile(os.path.join(entry.get_path(), 'stack.xml')):
        return entry
      else:
        return None
  return None

def roslocate_info(stack, distro, dev):
  """
  Looks up stack yaml on the web
  
  :raises: ROSInstallException on errors
  """
  # TODO: use roslocate from code
  cmd = ['roslocate', 'info', '--distro=%s'%(distro), stack]
  if dev == True:
    cmd.append('--dev')
  try:
    p = Popen(cmd, stdout=PIPE, stderr=PIPE)
  except OSError as e:
    raise ROSInstallException('%s\nfailed to execute roslocate; is your ROS environment configured?'%(e))

  stdout, stderr = p.communicate()
  if p.returncode != 0:
    sys.stderr.write('[rosws] Warning: failed to locate stack "%s" in distro "%s".  Falling back on non-distro-specific search; compatibility problems may ensue.\n'%(stack,distro))
    # Could be that the stack hasn't been released; try roslocate
    # again, without specifying the distro.
    cmd = ['roslocate', 'info', stack]
    if dev == True:
      cmd.append('--dev')
    try:
      p = Popen(cmd, stdout=PIPE, stderr=PIPE)
    except OSError as e:
      raise ROSInstallException('%s\nfailed to execute roslocate; is your ROS environment configured?'%(e))

    stdout, stderr = p.communicate()
    if p.returncode != 0:
      raise ROSInstallException('roslocate failed: %s'%(stderr))
  return yaml.load(stdout)


def get_ros_stack_version():
  """
  Reads/Infers the ros stack version. Avoid using this function if you can.
  """
    # TODO: switch to `rosversion -d` after it's been released (r14279,
  # r14280)
  cmd = ['rosversion', 'ros']
  try:
    p = Popen(cmd, stdout=PIPE, stderr=PIPE)
  except OSError as e:
    raise ROSInstallException('%s\nfailed to execute rosversion; is your ROS environment configured?'%(e))

  stdout, stderr = p.communicate()
  if p.returncode != 0:
    raise ROSInstallException('rosversion failed: %s'%(stderr))
  ver = distutils.version.StrictVersion(stdout).version
  if len(ver) < 2:
    raise ROSInstallException('invalid ros version: %s'%(stdout))
  return ver

def rosversion_to_distro_name(ver):
  """
  Reads/Infers the distro name from ROS / or the ros stack
  version. Avoid using this function if you can.
  """
  if len(ver) < 2:
    raise ROSInstallException('invalid ros version: %s'%(ver))
  major, minor = ver[0:2]
  if major == 1 and minor == 10:
    return 'groovy'
  if major == 1 and minor == 8:
    return 'fuerte'
  if major == 1 and minor == 6:
    return 'electric'
  elif major == 1 and minor == 5:
    return 'unstable'
  elif major == 1 and minor == 4:
    return 'diamondback'
  else:
    raise ROSInstallException('unknown ros version: %s'%(ver))

def get_dependent_stacks(stack):
  """
  Calls rosstack depends-on to get a list of dependance stacks. Avoid
  using this function if you can.
  """
  # roslib.stacks doesn't expose the dependency parts of rosstack, so
  # we'll call it manually
  cmd = ['rosstack', 'depends-on', stack]
  try:
    p = Popen(cmd, stdout=PIPE, stderr=PIPE)
  except OSError as e:
    raise ROSInstallException('%s\nfailed to execute rosstack; is your ROS environment configured?'%(e))
  stdout, stderr = p.communicate()
  if p.returncode != 0:
    raise ROSInstallException('rosstack failed: %s'%(stderr))
  # Make sure to exclude empty lines
  deps = []
  for l in stdout.split('\n'):
    if len(l) > 0:
      deps.append(l)
  return deps



def cmd_add_stack(config, stackname, released = False, recurse = False):
  """
  Attempts to get ROS stack from source if it is not already in config.
  Attempts the same for all stacks it depents, if recurse is given.
  Fails if any stack failed.
  
  :param released: use the released or the dev version
  :param recurse: also get dependant version
  :returns: True if stack has been added
  """
  def _add_stack(config, stackname, distro, released = False, recurse = False):
    stack_element = get_stack_element_in_config(config, stackname)
    if stack_element is not None:
      print("stack %stackname already in config at %s"%(stackname, stack_element.get_path()))
      return False
    yaml_dict = roslocate_info(stackname, distro, not released)
    if yaml_dict is not None and len(yaml_dict) > 0:
      path_spec = rosinstall.config_yaml.get_path_spec_from_yaml(yaml_dict[0])
      if config.add_path_specs([path_spec]) == False:
        print("Config did not add element %s"%path_spec)
        return False
      return True
    print("roslocate did not return anything")
    return False

  ver = get_ros_stack_version()
  distro = rosversion_to_distro_name(ver)
  if _add_stack(config, stackname, distro, released, recurse) == False:
    return False
  
  if recurse:
    deps = get_dependent_stacks(stack)
    # Also switch anything that depends on this stack
    for s in deps:
      if add_stack_rec(config, s, distro, not released) == False:
        return False
  return True

def cmd_delete_stack(config, stackname, delete = False, recurse = False):
  """
  Attempts to get ROS stack from source if it is not already in config.
  Attempts the same for all stacks it depents, if recurse is given.
  Fails if any stack failed.
  
  :param released: use the released or the dev version
  :param recurse: also get dependant version
  :returns: True if stack has been added
  """
  def _del_stack(config, stackname, delete = False, recurse = False):
    stack_element = get_stack_element_in_config(config, stackname)
    if stack_element is None:
      print("stack not in config: %s "%stackname)
      return False
    config.remove_element(stack_element.get_local_name())
    if delete:
      # TODO confirm each delete
      shutil.rmtree(os.path.join(self.path, stack), ignore_errors=True)
    return True

  if _del_stack(config, stackname, delete, recurse) == False:
    return False
  
  if recurse:
    deps = get_dependent_stacks(stack)
    # Also switch anything that depends on this stack
    for s in deps:
      _del_stack(config, s, delete, recurse)
  return True



class RosWsStacksCLI():

    def __init__(self):
        self.config_filename = ROSINSTALL_FILENAME
  
    def cmd_add_stack(self, target_path, argv):
        parser = OptionParser(usage="usage: rosws add-stack [PATH] localname",
                        epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("-N", "--non-recursive", dest="norecurse", default=False,
                          help="don't change configuration for dependent stacks",
                          action="store_true")
        parser.add_option("--released", dest="released", default=False,
                          help="Pull stack from release tag instead of development branch",
                          action="store_true")
        parser.add_option("--continue-on-error", dest="robust", default=False,
                          help="Continue despite checkout errors",
                          action="store_true")
        parser.add_option("--delete-changed-uris", dest="delete_changed", default=False,
                          help="Delete the local copy of a directory before changing uri.",
                          action="store_true")
        parser.add_option("--abort-changed-uris", dest="abort_changed", default=False,
                          help="Abort if changed uri detected",
                          action="store_true")
        parser.add_option("--backup-changed-uris", dest="backup_changed", default='',
                          help="backup the local copy of a directory before changing uri to this directory.",
                          action="store")
        (options, args) = parser.parse_args(argv)
        mode = rosinstall.rosws_cli._get_mode_from_options(parser, options)
        if len(args) < 1:
            print("Error: Too few arguments.")
            print(parser.usage)
            return -1
        if len(args) > 1:
            print("Error: Too many arguments.")
            print(parser.usage)
            return -1
        stack = args[0]
        config = rosinstall.multiproject_cmd.get_config(target_path, [], config_filename = self.config_filename)
        if cmd_add_stack(config, stack) == True:
            rosinstall.multiproject_cmd.cmd_persist_config(config, self.config_filename)
            # install or update each element
            install_success = rosinstall.multiproject_cmd.cmd_install_or_update(config,
                                                                     backup_changed = options.backup_changed,
                                                                     mode = mode,
                                                                     robust = options.robust)
            if install_success:
              return 0
        return 1


    def cmd_delete_stack(self, target_path, argv):
        parser = OptionParser(usage="usage: rosws delete-stack [PATH] localname",
                        epilog="See: http://www.ros.org/wiki/rosinstall for details\n")
        parser.add_option("-N", "--non-recursive", dest="norecurse", default=False,
                          help="don't change configuration for dependent stacks",
                          action="store_true")
        parser.add_option("-d", "--delete-working-copies", dest="delete", default=False,
                          help="when deleting a stack from the configuration, also delete the working copy (DANGEROUS!)",
                          action="store_true")
        (options, args) = parser.parse_args(argv)

        if len(args) < 1:
            print("Error: Too few arguments.")
            print(parser.usage)
            return -1
        if len(args) > 1:
            print("Error: Too many arguments.")
            print(parser.usage)
            return -1
        uri = args[0]
        config = rosinstall.multiproject_cmd.get_config(target_path, [], config_filename = self.config_filename)
        if cmd_delete_stack(config, uri):
            rosinstall.multiproject_cmd.cmd_persist_config(config, self.config_filename)
            return 0
        return 1
    

  

def usage():
  print("""%(prog)s is an experimental command to add and remove stack from ROS workspaces.

Usage:
  %(prog)s add [INSTALL_PATH] [STACK] [OPTIONS]
  %(prog)s delete [INSTALL_PATH] [STACK] [OPTIONS]

Type '%(prog)s --help' for usage.
"""%{'prog': 'rosws-stacks'})
    
def rosws_stacks_main(argv=None):
  """
  Calls the function corresponding to the first argument.
  """
  if argv == None:
      argv = sys.argv
  if ('--help' in argv):
      usage()
      return 0
  if len(argv) < 2:
      try:
          workspace = rosinstall.cli_common.get_workspace(argv, os.getcwd(), config_filename = ROSINSTALL_FILENAME, varname = "ROS_WORKSPACE")
          argv.append('info')
      except MultiProjectException as e:
        print(str(e))
        usage()
        return 0


  try:
    command = argv[1]
    args = argv[2:]

    if command == 'help':
      if len(argv) < 3:
        usage()
        return 0

      else:
        command=argv[2]
        args = argv[3:]
        args.insert(0,"-h")

    cli = RosWsStacksCLI()
    commands = {
      'add'         : cli.cmd_add_stack,
      'delete'      : cli.cmd_delete_stack,
      }
    if command not in commands:
      if os.path.exists(command):
        args = ['-t', command] + args
        command = 'info'
      else:
        if command.startswith('-'):
          print("First argument must be name of a command: %s"%command)
        else:
          print("Error: unknown command: %s"%command)
        usage()
        return 1
    workspace = rosinstall.cli_common.get_workspace(args, os.getcwd(), config_filename = ROSINSTALL_FILENAME)
    result = commands[command](workspace, args) or 0
    return result

  except KeyboardInterrupt:
    pass



# Software License Agreement (BSD License)
#
# Copyright (c) 2009, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of Willow Garage, Inc. nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
import os
from config_elements import AVCSConfigElement, OtherConfigElement, SetupConfigElement
from common import MultiProjectException, normabspath, realpath_relation, is_web_uri


class Config:
  """
  A config is a set of config elements, each of which defines a folder or file
  and possibly a VCS from which to update the folder.
  """
  
  def __init__(self, path_specs, install_path, config_filename = None, extended_types=None, merge_strategy = 'KillAppend'):
    """
    :param config_source_dict: A list (e.g. from yaml) describing the config, list of dict, each dict describing one element.
    :param config_filename: When given a folder, Config
    :param merge_strategy: how to deal with entries with equivalent path. See insert_element
    
    will look in folder for file of that name for more config source, str.
    """
    assert install_path is not None
    if path_specs is None:
      raise MultiProjectException("Passed empty source to create config")
    # All API operations must grant that elements in trees have unique local_name and paths
    # Also managed (VCS) entries must be disjunct (meaning one cannot be in a child folder of another managed one)
    # The idea is that managed entries can safely be concurrently modified
    self.trees = []
    self.base_path = os.path.abspath(install_path)

    self.config_filename = None
    if config_filename is not None:
      self.config_filename = os.path.basename(config_filename)
    # using a registry primarily for unit test design
    self.registry = {'svn': AVCSConfigElement,
                     'git': AVCSConfigElement,
                     'hg': AVCSConfigElement,
                     'bzr': AVCSConfigElement,
                     'tar': AVCSConfigElement}
    if extended_types is not None:
      self.registry = dict(list(self.registry.items()) + list(extended_types.items()))

    for path_spec in path_specs:
      action = self.add_path_spec(path_spec, merge_strategy)
      # Usual action in init should be 'Append', anything else is unusual
      if action == 'KillAppend':
        print("Replace existing entry %s by appending."%path_spec.get_local_name())
      elif action == 'MergeReplace':
        print("Replace existing entry %s"%path_spec.get_local_name())
      elif action == 'MergeKeep':
        print("Keep existing entry %s, discard later one"%path_spec.get_local_name())

  def __str__(self):
    return str([str(x) for x in self.trees])

  def add_path_spec(self, path_spec, merge_strategy = 'KillAppend'):
    """
   goes through path_specs and builds up self.trees.
   May recursively pull elements from remote sources.
   
   :param merge_strategy: see insert_element
   :param path_specs: PathSpec objects
   :returns: merge action taken, see insert_element
   """
    #compute the local_path for the config element
    local_path = normabspath(path_spec.get_local_name(), self.get_base_path())
    if path_spec.get_scmtype() != None:
      return self._insert_vcs_path_spec(path_spec, local_path, merge_strategy)
    else:
      if path_spec.get_tags() is not None and 'setup-file' in path_spec.get_tags():
        elem = SetupConfigElement(local_path, os.path.normpath(path_spec.get_local_name()), properties = path_spec.get_tags())
        return self.insert_element(elem, merge_strategy)
      else:
        # we dont want files or other Config folders in this Config
        if not (os.path.isfile(local_path)
                or
                (self.config_filename is not None
                 and os.path.isdir(local_path)
                 and os.path.exists(os.path.join(local_path, self.config_filename)))):
          local_name = os.path.normpath(path_spec.get_local_name())
          elem = OtherConfigElement(local_path,
                                    local_name,
                                    path_spec.get_uri(),
                                    path_spec.get_version(),
                                    properties = path_spec.get_tags())
          return self.insert_element(elem, merge_strategy)
        else:
          print("!!!!!Warning: Not recursing into other config folder %s"%local_path)


  def _insert_vcs_path_spec(self, path_spec, local_path, merge_strategy = 'KillAppend'):
    # Get the version and source_uri elements
    source_uri = path_spec.get_uri()
    
    if (source_uri is not None
        and not is_web_uri(source_uri)
        and not os.path.isabs(source_uri)):
      source_uri = os.path.normpath(os.path.join(self.get_base_path(), source_uri))
      print("Warning: Converted relative uri path %s to abspath %s"%(path_spec.get_uri(), source_uri))
    version = path_spec.get_version()
    try:
      local_name = os.path.normpath(path_spec.get_local_name())
      elem = self._create_vcs_config_element(path_spec.get_scmtype(),
                                             local_path,
                                             local_name,
                                             source_uri,
                                             version,
                                             properties = path_spec.get_tags())
      return self.insert_element(elem, merge_strategy)
    except LookupError as ex:
      raise MultiProjectException("Abstracted VCS Config failed. Exception: %s" % ex)

  def insert_element(self, new_config_elt, merge_strategy = 'KillAppend'):
    """
   Insert ConfigElement to self.trees, checking for duplicate local-name or path first.
   In case local_name matches, follow given strategy
   
   - KillAppend (default): remove old element, append new at the end
   - MergeReplace: remove first such old element, insert new at that position.
   - MergeKeep: Discard new element
   
   In case local path matches but local name does not, raise Exception
   
   :returns: the action performed None, 'Append', 'KillAppend', 'MergeReplace', 'MergeKeep'"""
    removals = []
    replaced = False
    for index, loop_elt in enumerate (self.trees):
      # if paths are os.path.realpath, no symlink problems.
      relationship = realpath_relation(loop_elt.get_path(), new_config_elt.get_path())
      if relationship == 'SAME_AS':
        if os.path.normpath(loop_elt.get_local_name()) != os.path.normpath(new_config_elt.get_local_name()):
          raise MultiProjectException("Elements with different local_name target the same path: %s, %s"%(loop_elt, new_config_elt))
        else:
          if (loop_elt == new_config_elt):
            return None
          if (merge_strategy == 'MergeReplace'
              or (merge_strategy == 'KillAppend' and index == len(self.trees)-1)):
            self.trees[index] = new_config_elt
            # keep looping to check for overlap when replacing non-scm with scm entry
            replaced = True
            if (loop_elt.is_vcs_element or not new_config_elt.is_vcs_element):
              return 'MergeReplace'
          elif merge_strategy == 'KillAppend':
            removals.append(loop_elt)
          elif merge_strategy == 'MergeKeep':
            return 'MergeKeep'
          else:
            raise LookupError("No such merge strategy: %s"%str(merge_strategy))
      elif ((relationship == 'CHILD_OF' and new_config_elt.is_vcs_element())
            or (relationship == 'PARENT_OF' and loop_elt.is_vcs_element())):
        # we do not allow any elements to be children of scm elements
        # to allow for parallel updates and because rosinstall may
        # delete scm folders on update, and thus subfolders can be
        # deleted with their parents
        raise MultiProjectException("Managed Element paths overlap: %s, %s"%(loop_elt, new_config_elt))
    if replaced:
      return 'MergeReplace'
    for loop_elt in removals:
      self.trees.remove(loop_elt)
    self.trees.append(new_config_elt)
    if len(removals) > 0:
      return 'KillAppend'
    return 'Append'

  def remove_element(self, local_name):
    """
    Removes element in the tree with the given local name (should be only one)
    
    :returns: True if such an element was found
    """
    removals = []
    for t in self.trees:
      if t.get_local_name() == local_name:
        removals.append(t)
    if len(removals) > 0:
      for t in removals:
        self.trees.remove(t)
      return True
    return False

  def _create_vcs_config_element(self, scmtype, path, local_name, uri, version = '', properties = None):
    try:
      eclass = self.registry[scmtype]
    except LookupError:
      raise MultiProjectException("No VCS client registered for vcs type %s"%scmtype)
    return eclass(scmtype = scmtype, path = path, local_name = local_name, uri = uri, version = version, properties = properties)

  def get_base_path(self):
    return self.base_path

  def get_config_filename(self):
    return self.config_filename
  

  def get_source(self):
    """
    :returns: all elements that got added by user keystrokes (CLI and changed .rosinstall)
    """
    source_aggregate = []
    for t in self.trees:
      source_aggregate.append(t.get_path_spec())
    return source_aggregate  

  
  def get_config_elements(self):
    source_aggregate = []
    for t in self.trees:
      source_aggregate.append(t)
    return source_aggregate

